# Copyright (c) 2011, Chandler Armstrong (omni dot armstrong at gmail dot com)
# see LICENSE.txt for details



from collections import deque
from pygame import Rect, Surface
from pygame import font
from pygame.font import Font
from pygame.sprite import Sprite
from pygame.locals import *
import re



######################################################################
# inits
font.init()



######################################################################
# constants
DIRECTION_KEYS = [K_UP, K_DOWN, K_LEFT, K_RIGHT]
BLACK = (0, 0, 0)
WHITE = (255, 255, 255)
FONT = Font(None, 8)



######################################################################
# functions
def _iswhitespace(char):
    # determine if char is whitespace
    # accepts a char, if char is not a string then it is always
    # considered whitespace
    # returns true if char is whitespace, else returns false
    whitespace = re.compile('\s+') # re to detect whitespace
    if isinstance(char, str):
        assert len(char) == 1
        if whitespace.search(char): return True
        return False
    return False



class _Line(object):


    def __init__(self, image, rect):
        self.image = image
        self.rect = rect


    def clear(self, surface_dest, background):
        # draws background over surface_dest using self.rect
        # accepts surface_dest: the destination surface that self.image has
        # been drawn to
        # background: the background to draw over the areas that self.image was
        # drawn to on surface_dest
        # returns nothing
        rect = self.rect
        self.image = Surface(rect.size)
        surface_dest.blit(background, rect, rect)



######################################################################
class EditorGroup(dict):
    """a dictionary that assumes editors as values.  adds methods to retrieve
    editors given a position"""


    def get_focus(self, pos):
        for k, v in self.items():
            if v.rect.collidepoint(pos): return k, v
        return None, None



class Editor(object):


    def __init__(self, rect, txt=None,
                 bkg=BLACK, color=WHITE, font=FONT, spacing=0):
        self.image = Surface(rect.size)
        self._image = Surface(rect.size)
        self.rect = rect
        if txt is None: self.txt = [' ']
        else: self.txt = txt
        self._lines = [_Line(Surface((0, 0)), Rect(0, 0, 0, 0))]
        self._wraps = [0]
        self._cursor = 0
        self._dest = Rect(0, 0, 0, 0)
        self.bkg = bkg
        self.color = color
        self.font = font
        self.spacing = spacing


    def get_cursor(self):
        """
        get a rect representing the position of the cursor on the editor image
        """
        k, font, txt = self._cursor, self.font, self.txt
        index2pixel = self._index2pixel

        return Rect(index2pixel(k), font.size(txt[k]))


    def clear(self):
        """clear the editor image of all lines and reset the cursor"""
        lines = self._lines
        image, bkg_image = self.image, self._image
        for line in lines: line.clear(image, bkg_img)
        self._cursor = 0


    def _index2line(self, i):
        wraps = self._wraps

        for l, wrap in enumerate(wraps):
            if wrap > i: return l - 1

        return len(wraps) - 1


    def _index2pixel(self, i):
        font, lines, txt = self.font, self._lines, self.txt
        wraps = self._wraps + [len(txt)]
        index2line = self._index2line

        l = index2line(i)

        x = font.size(''.join(txt[wraps[l]:i]))[0]
        y = lines[l].rect.y

        return x, y


    def _pixel2index(self, pixel):
        font, txt, wraps = self.font, self.txt, self._wraps
        pixel2line = self._pixel2line

        l = pixel2line(pixel)
        if l == len(wraps) - 1: m, n = wraps[l], len(txt)

        else: m, n = wraps[l:l + 2]

        pixel_x = pixel[0]
        d = deque([0, 0], 2)
        for i in xrange(m, n):
            font_size_x = font.size(''.join(txt[m:i]))[0]
            d.append(font_size_x)
            if font_size_x > pixel_x: break

        l = [abs(font_size_x - pixel_x) for font_size_x in d]

        if l.index(min(l)) == 1: return i

        return i - 1


    def _pixel2line(self, pixel):
        lines = self._lines

        for l, line in enumerate(lines):
            if line.rect.y > pixel[1]: return l - 1

        return l


    def _draw_line(self, l):
        image, lines = self.image, self._lines

        line = lines[l]
        image.blit(line.image, line.rect)


    def redraw(self):
        image, lines = self.image, self._lines
        for line in lines: image.blit(line.image, line.rect)


    def input(self, event):
        """
        accept a user KEYDOWN event and add the text associated with the
        keypress to the editor image

        event-- a user KEYDOWN event

        return None
        """
        k, font, lines = self._cursor, self.font, self._lines
        rect_w, txt, wraps = self.rect.w, self.txt, self._wraps
        image, bkg_img = self.image, self._image
        draw_line, index2line = self._draw_line, self._index2line
        pixel2index, update = self._pixel2index, self._update
        Line = _Line

        if event.type == KEYDOWN:
            l = index2line(k)

            # handle cursor navigation
            if event.key in DIRECTION_KEYS:
                draw_line(l) #to clear the cursor?
                if event.key == K_UP:
                    if l > 0:
                        pixel = (font.size(''.join(txt[wraps[l]:k]))[0],
                                 lines[l - 1].rect.y)
                        self._cursor = pixel2index(pixel)

                    else: self._cursor = 0

                elif event.key == K_DOWN:
                    if (len(wraps) - 1) > l:
                        pixel = (font.size(''.join(txt[wraps[l]:k]))[0],
                                 lines[l + 1].rect.y)
                        self._cursor = pixel2index(pixel)

                    else: self._cursor = len(txt) - 1

                elif event.key == K_LEFT:
                    self._cursor -= 1
                    if self._cursor < 0: self._cursor = 0

                elif event.key == K_RIGHT:
                    self._cursor += 1
                    if self._cursor > (len(txt) - 1):
                        self._cursor = len(txt) - 1

            # handle newlines
            elif event.key == K_RETURN:
                txt.insert(k, '\n')
                self._cursor += 1
                lines.insert(l + 1, Line(Surface((0, 0)), Rect(0, 0, 0, 0)))
                update(l)

            # handle backspaces
            elif event.key == K_BACKSPACE:
                if k == 0: pass

                else:
                    lines[l].clear(image, bkg_img)
                    self._cursor -= 1
                    k = self._cursor
                    char = txt.pop(k)
                    # if l> index2line(k) then the line no longer exists
                    if l > index2line(k): del lines[l]
                    if l == 0: update(l)
                    else: update(l - 1)

            # handle ascii input
            else:
                if event.unicode:
                    txt.insert(k, event.unicode.encode("ascii"))
                    self._cursor += 1
                    if k == wraps[l]: # wrapped to a new line
                        if l == 0: update(l) # update line
                        else: update(l - 1)  # else: update from previous line
                    else: update(l)


    ##################################################################
    def _tokenize(self, txt):
        charbuff = []
        for char in txt:
            if _iswhitespace(char):
                if charbuff: yield ''.join(charbuff)
                charbuff = []
                yield char

            else: charbuff.append(char)

        if charbuff: yield ''.join(charbuff)


    def _update(self, l):
        bkg, color, font = self.bkg, self.color, self.font
        spacing, dest, rect_w = self.spacing, self._dest, self.rect.w
        bkg_img, image = self._image, self.image
        lines, txt, wraps = self._lines, self.txt, self._wraps
        tokenize = self._tokenize
        Line = _Line

        tokens = tokenize(txt[wraps[l]:])
        dest.y = sum(line.image.get_height() + spacing for line in lines[:l])
        wraps = wraps[:l + 1]
        k = wraps[-1]
        linebuff = []
        L = [l] # a list of updated lines
        for token in tokens:
            if token == '\n':
                k += 1 # increment k (so the \n will be on this line)
                wraps.append(k)
                # a space is the visual representation of the \n character
                linebuff.append(' ')
                _linebuff = []

            elif font.size(''.join(linebuff + [token]))[0] < rect_w:
                k += len(token)
                linebuff.append(token)
                continue

            else:
                wraps.append(k) # k is the position of the previous space token
                k += len(token)
                _linebuff = [token]
                lines.insert(l + 1, Line(Surface((0, 0)), Rect(0, 0, 0, 0)))

            line = font.render(''.join(linebuff), 0, color, bkg)
            line = Line(line, Rect(dest.topleft, line.get_size()))
            lines[l].clear(image, bkg_img)
            lines[l] = line

            dest.y += line.image.get_height() + spacing
            linebuff = _linebuff
            l += 1
            L.append(l)

        line = font.render(''.join(linebuff), 0, color, bkg)
        line = Line(line, Rect(dest.topleft, line.get_size()))
        lines[l].clear(image, bkg_img)
        lines[l] = line

        for line in [lines[l] for l in L]: image.blit(line.image, line.rect)

        self._wraps = wraps



# Copyright (c) 2011, Chandler Armstrong (omni dot armstrong at gmail dot com)
# see LICENSE.txt for details




from __future__ import division
import re
from collections import defaultdict, deque
import os
from sys import stderr

import pygame
from pygame import font
from pygame.event import Event
from pygame.font import Font
from pygame.rect import Rect
from pygame.sprite import Sprite
from pygame import Surface
from pygame.transform import scale
from pygame.locals import *

from .editor import Editor




######################################################################
# inits
font.init()




######################################################################
# constants
BLACK = (0, 0, 0)
WHITE = (255, 255, 255)
FONT = Font(None, 8)
SPECIALS = ['/', '{', '}']
WHITESPACE = {'n' : '\n'}
WARN_BUFF = '''\
Warning: buffer not emptied, try increasing rect height or rect width and add
more columns\n'''




######################################################################
# globals
Macros = {}




######################################################################
# private functions
def _iswhitespace(char):
    # determine if char is whitespace
    # accepts a char, if char is not a string then it is never
    # considered whitespace (to prevent images from being stretched)
    # returns true if char is whitespace, else returns false
    whitespace = re.compile('\s+') # re to detect whitespace
    if not char: return True
    if isinstance(char, str) or isinstance(char, unicode):
        assert len(char) == 1
        if whitespace.search(char): return True
        return False
    return False



def _scale_token(token, (width, height)):
    # scale token to width and height
    # accepts token object and a (width, height) argument
    # returns a new token object scaled to width and height
    Token = _Token

    links = token.links
    token_iswhitespace, token_str = token.iswhitespace, token.str
    surfs = [scale(token, (width, height))]

    ratio_w = width / token.get_width()
    ratio_h = height / token.get_height()

    # given token height, modify link rect y
    shift_x = 0
    for rect in [rect for v in links.values() for rect in v]:
        rect_w = rect.w
        rect.w *= ratio_w
        rect.h *= ratio_h
        rect.x += shift_x
        rect.y += (height - rect.h)    
        shift_x = rect.w - rect_w

    return Token((width, height), links, surfs, token_iswhitespace, token_str)



def _token_builder(interpreted_txt):
    # build a token from interpreted text
    # accepts an interpreted text list
    # returns a token object
    iswhitespace = _iswhitespace
    Token = _Token

    links = defaultdict(list)

    token_iswhitespace = True

    surfs, x, y = [], 0, 0
    for (envs, chars) in interpreted_txt:
        bkg, color, font = envs['bkg'], envs['color'], envs['font']
        strbuff, surfbuff = [], []
        for char in chars:
            if not iswhitespace(char): token_iswhitespace = False

            if char == '\n':
              char = Surface((0, font.get_linesize()))

            if isinstance(char, unicode):
                if iswhitespace(char): char = ' '
                strbuff.append(char)
            elif isinstance(char, str):
                if iswhitespace(char): char = ' '
                strbuff.append(char)            
            else:
                if strbuff:
                  surfbuff.append(font.render(''.join(strbuff), 1, color, bkg))
                surfbuff.append(char)
                strbuff = []
        
        if strbuff:
          surfbuff.append(font.render(''.join(strbuff), 1, color, bkg))

        if surfbuff:
            # calculate link rects
            link = envs['link']
            surfbuff_w = sum(surf.get_width() for surf in surfbuff)
            surfbuff_h = max(surf.get_height() for surf in surfbuff)
            links[link].append(Rect(x, 0, surfbuff_w, surfbuff_h))
            x += surfbuff_w
            # extend surfbuff to surfs and reset surfbuff
            surfs.extend(surfbuff)
            surfbuff = []

    # get token width and height
    width = sum(surf.get_width() for surf in surfs)
    height = max(surf.get_height() for surf in surfs)

    # given token height, modify link rect y
    for rect in [rect for v in links.values() for rect in v]:
        rect.y += (height - rect.h)
    token_str = ''.join(unicode(char) for (envs, chars) in interpreted_txt
                        for char in chars)

    return Token((width, height), links, surfs, token_iswhitespace, token_str)




######################################################################
# private classes
class _Token(Surface):
    # token object
    # links is a dictionary of link id strings keyed to the link rect on the
    #   token
    # iswhitespace is a boolean indicating if the token is whitespace
    # str is a string representing the token content


    def __init__(self, (width, height), links, surfs, iswhitespace, token_str):
        # construct a token object
        # (width, height) is the width and height of the token
        # links is a dictionary of link id strings keyed to the link rect on the
        #   token
        # surfs is a list of surfaces comprising the token
        # iswhitespace is a boolean indicating if the token is whitespace
        # token_str is a string representing the token
        self.links = links
        self.iswhitespace = iswhitespace
        self.str = token_str

        # construct Surface superclass
        Surface.__init__(self, (width, height))

        # blit each token onto self
        x = 0
        for surf in surfs:
            w, h = surf.get_size()
            y = height - h
            self.blit(surf, (x, y))
            x += w


    def __str__(self):
        return self.str



class _Line(Surface):
    # Line class


    def __init__(self, line, surf_w, justify):
        # renders a line (list of tokens)
        # accepts a line (list of tokens), width of line, and justification
        # returns Surface object with the tokens justified upon it given the
        #   justify argument
        self.line = line
        self.links = defaultdict(list)

        links = self.links
        scale_token = _scale_token

        # get the line width, line height, whitespace used in the line, and
        # freespace remaining in the line
        line_w = sum(token.get_width() for token in line)
        ### FUTURE center
        # line_cw = sum(token.get_width() for token in line if not token.center)
        # iine_w - line_cw
        # if line_cw: line_w = line_cw
        line_h = max(token.get_height() for token in line)
        total_whitespace = sum(token.get_width() for token in line
                               if token.iswhitespace)
        freespace = surf_w - line_w # the freespace available in the surface

        # initialize surface
        Surface.__init__(self, (surf_w, line_h))
        self.set_colorkey(BLACK)

        # set x
        if justify == 'right': x = freespace
        elif justify == 'center': x = freespace / 2
        else: x = 0

        # if justified: modify whitespace widths
        if justify == 'justified':
            # the amount to stretch whitespace
            if total_whitespace: stretch = 1.0 + (freespace / total_whitespace)
            # if not total_whitespace, stretch will not be used.

            sub_w = 0
            for i, token in enumerate(line):
                if token.iswhitespace:
                    # replace token with a scaled token
                    token_w = token.get_width() * stretch
                    scale_w = int(token_w)
                    sub_w += token_w % scale_w
                    while sub_w > 1:
                        sub_w -= 1
                        scale_w += 1
                    line[i] = scale_token(token, (scale_w, token.get_height()))

        for token in line:
            w, h = token.get_size()
            y = (line_h - h) # token y
            self.blit(token, (x, y))

            for link in token.links:
                for rect in token.links[link]:
                    # move rect to token's position in line and append to links
                    links[link].append(rect.move(x, y))

            x += w # update x with object width


    def __str__(self):
        return ''.join(str(token) for token in self.line)



######################################################################
# public classes
class GlyphGroup(dict):
    """a dictionary that assumes Glyphs as values.  adds methods to get links
    from all Glyphs in the group"""

    def get_collisions(self, pos):
        for k, v in self.items():
            link =  v.get_collisions(pos)
            if link: return k, link
        return None, None

    def draw(self, surface):
        for v in self.values(): surface.blit(v.image, v.rect)

    def update(self):
        for v in self.values(): v.update()


class Glyph(object):
    """
    Main glyph class
    image-- the image that text is set to
    rect-- rect for blitting image to viewing surface
    spacing-- line spacing
    links-- dict of (link, [rects]) pairs
    """


    ##################################################################
    # class methods
    def __init__(self, rect, bkg=BLACK, color=WHITE, font=FONT, spacing=0,
                 ncols=1, col_space=20): # FUTURE COLS ADD
        """
        Initialize a glyph object

        rect-- rect object for positioning glyph image on viewing surface
        **kwargs-- dictionary of (env_id, value) pairs
          required kwargs are:
          bkg-- background color
          color-- font color
          font-- font
        """
        # initialize
        self.image = Surface(rect.size)
        self.image.fill(bkg)
        self.image.set_alpha(255)
        self._bkg = bkg
        self.rect = rect
        self.spacing = spacing
        self.links = defaultdict(list)
        # FUTURE ###
        self.editors = {}
        ############
        # FUTURE COLS ###
        self.col_w = ((rect.w + col_space) / ncols) - col_space
        self.col_space = col_space
        self.col_n = 1
        self.ncols = ncols
        #################
        self._dest = Rect(0, 0, 0, 0) # rect to blit a txt line to image surface
        # list of (env_id, value) pairs for environments;
        # _envs is used as a stack data structure
        self._envs = [('bkg', bkg),
                      ('color', color),
                      ('font', font),
                      ('link', None)] # link id string

        self.buff = deque() # rendered text buffer


    ##################################################################
    # helper methods
    def __read_env(self, _txt_):
        # interprets and returns an environment.  environments set text
        # characteristics such as color or links.
        # accepts the _txt_ iterable at an environment starting point
        # return (environment type, environment) tuple (e.g (font, Font object))

        # re to get env arguments (arguments may be paths and contain \ or /)
        r = re.compile('(\w+)(\s+((\"|\').*?(\"|\')|.*?))?;')
        charbuffer = ''
        # _txt_ is a generator, so iterating consumes the contents for the
        # references to _txt_ in the _interpret function
        for i, char in _txt_:
            charbuffer += char
            s = r.search(charbuffer) # search for environment name arguments
            if s: # if search successful
                groups = s.groups() # get environment
                env, args = groups[0], groups[2]
                if env in Macros: return Macros[env]
                # new environment types must be added here

                elif env == 'bkg':
                    # return new backgroun color
                    return ('bkg', tuple([int(e) for e in args.split(',')]))

                elif env == 'color':
                    # return new font color
                    return ('color', tuple([int(e) for e in args.split(',')]))

                elif env == 'font':
                    # return new font
                    path, size = args.split(',') # the font location and size
                    return ('font', Font(os.path.realpath(path), int(size)))

                elif env == 'link':
                    # return new link
                    return ('link', args.strip())

                # FUTURE ###
                elif env == 'editor':
                    #editor is considered an environment because it must be
                    #linked.  any text in an editor environment is input to
                    #that editor, and any nested environments are ignored.
                    name, w = args.split(',')
                    #extract editor kw args
                    kw = dict(self._envs)
                    del kw['link']
                    kw['spacing'] = self.spacing
                    h = kw['font'].get_linesize()
                    editor = Editor(Rect(0, 0, int(w), h), **kw)
                    self.editors[name] = editor 
                    # treat as link env, get_collision will sort 
                    return ('link', name)
                ############

                else:
                    raise ValueError(env + ' is an unrecognized environment')


    # the space func could take a size and return a surface or rect...
    # really only need to shift tokens.  surfs do that without requiring tokens
    # to have rects
    def __read_func(self, _txt_):
        # interprets and returns a function.  functions are special surfaces or
        # objects.
        # accepts the _txt_ iterable at function starting point
        # returns (function type, function results) tuple
        #   (eg. (space, Surface obejct))
        r = re.compile('(\w+){(.*?)}')
        i, char = _txt_.next()

        if char in SPECIALS: return 'special', char
        if char in WHITESPACE: return 'whitespace', WHITESPACE[char]

        charbuff = char
        for i, char in _txt_:
            charbuff += char
            s = r.search(charbuff)
            if s:
                func, args = s.groups()
                if func in Macros: return func, Macros[func]

                if func == 'space': return func, Surface((int(args), 1))
                #"if charbuff = 'img'"?
                if func == 'img': return func, pygame.image.load(args).convert()


    ##################################################################
    # private methods
    def _interpret(self, txt):
        # iterprets glyph markup language
        # accepts string literal
        # returns a list of (env, charbuff) pairs,
        #   where env is a dictionary of environment types keyed to values and
        #   charbuff a list of text strings and the surfaces created from
        #   functions
        editors, envs = self.editors, self._envs
        read_env, read_func = self.__read_env, self.__read_func
        iswhitespace = _iswhitespace

        txt = txt.strip()
        # FUTURE ###
        # preamble, txt = read_preamble(txt)
        # if preamble: envs = preamble
        # ##########

        # initialize charbuff, renderbuff, interpreted text, and previous char
        charbuff, interpreted_txt, prevchar = [], [], ''
        _txt_ = enumerate(txt)
        for i, char in _txt_:
            if iswhitespace(char) and iswhitespace(prevchar):
                if char == '\n': charbuff[-1] = char
                continue

            if char == '/': # a function:
                func, char = read_func(_txt_)
                charbuff.append(char)
                prevchar = char

            elif char == '{': # a new environment has started
                # using dict(envs) allows new environments to overwrite default
                # environments, which are in the beginning of the list
                interpreted_txt.append((dict(envs), charbuff))
                charbuff = []
                envs.append(read_env(_txt_))

            elif char == '}': # an environment has ended
                # FUTURE ###
                link = dict(envs)['link']
                if link in editors:
                    editor = editors[link]
                    # this is a hack to feed char to editor using input method
                    for char in charbuff:
                        mod = 0
                        if char.isupper(): mod = 3
                        event = Event(KEYDOWN, key=None, mod=mod,
                                      unicode=char.encode('utf8'))
                        editor.input(event)
                    interpreted_txt.append((dict(envs), [editor.image]))
                else: interpreted_txt.append((dict(envs), charbuff))
                #interpreted_txt.append((dict(envs), charbuff)) # FUTURE DEL
                ############
                charbuff = []
                envs.pop()

            else: # a normal, string, character
                charbuff.append(char)
                prevchar = char
        if charbuff: interpreted_txt.append((dict(envs), charbuff))
        return interpreted_txt


    def _tokenize(self, interpreted_txt):
        # tokenizes text
        # accepts (envs, charbuff)
        # returns a list of tokens
        iswhitespace, token_builder = _iswhitespace, _token_builder

        charbuff, _interpreted_txt, tokenized_txt = [], [], []
        for (envs, chars) in interpreted_txt:
            for char in chars:
                if iswhitespace(char):
                    if charbuff:
                        _interpreted_txt.append((envs, charbuff))
                        charbuff = []

                    if _interpreted_txt:
                        yield token_builder(_interpreted_txt)
                        _interpreted_txt = []

                    yield token_builder([(envs, [char])])

                else: charbuff.append(char)

            if charbuff:
                _interpreted_txt.append((envs, charbuff))
                charbuff = []
        if _interpreted_txt: yield token_builder(_interpreted_txt)


    def _wrap(self, tokenized_txt, justify):
        # wrap interpreted text to page width
        # accepts a list of Token objects tuples
        # returns a list of Line objects, each wrapped to self.rect
        Line = _Line
        # FUTURE COLS ###
        rect_w = self.col_w
        #rect_w = self.rect.w # FUTURE COLS DEL
        #################

        # linesize tracks the current size of the rendered line because moving
        # between environments will mean that there will be multiple surfaces
        # that need to be glued together
        line = [] # initialize line, and linesize
        for token in tokenized_txt:
            token_w = token.get_width()
            if token_w > rect_w:
                raise ValueError("The token '"+token.str+"' is "
                                 +str((token_w - rect_w))
                                 +" pixels too wide to fit in the rect passed.")

            # the token fits, process it and check if the line still fits inside
            # rect area, if not, append line without token, reinitialize line
            # with token
            line.append(token)
            if unicode(token) == '\n':
                # don't justify a line that would not wrap
                if justify == 'justified': _justify = 'left'
                else: _justify = justify

                yield Line(line, rect_w, _justify)
                line = [] # reset line

            elif sum(token.get_width() for token in line) > rect_w:
                token = line.pop()
                # remove single trailing whitespace
                if line[-1].iswhitespace: line = line[:-1]

                yield Line(line, rect_w, justify)
                line = [] # reinitialize _line, line, linesize, and num objects
                # do not append whitespace as the first token of the new line
                if not token.iswhitespace: line.append(token)

        if line:
            # don't justify a line that would not wrap
            if justify == 'justified': _justify = 'left'
            else: _justify = justify

            yield Line(line, rect_w, _justify)


    ##################################################################
    # public methods
    def input(self, txt, justify=None, update=True):
        """
        interprets, renders, wraps, and justifies input text

        txt -- raw text written with glyph markup
        justify -- a justify command; default is left justified
          left: left justified
          right: right justified
          justified: justified (both sides even)
          center: center justified

        returns nothing
        """
        buff, interpret, tokenize, wrap = (self.buff, self._interpret,
                                           self._tokenize, self._wrap)
        interpreted_txt = interpret(txt)
        tokens = tokenize(interpreted_txt)
        lines = wrap(tokens, justify)
        buff.extend(lines)
        if update: self.update()


    def overwrite(self, txt, **kw):
        self.clear()
        self.input(txt, **kw)


    def update(self):
        """
        updates the surface with the text input to the buffer by the input
        method, then deletes buffer

        accepts nothing

        returns nothing
        """
        buff, dest = self.buff, self._dest
        spacing = self.spacing
        image, rect = self.image, self.rect
        editors, links = self.editors, self.links
        # FUTURE COLS ###
        ncols, col_n = self.ncols, self.col_n
        col_w, col_space = self.col_w, self.col_space
        #################

        while buff:
            line = buff.popleft()
            line_h = line.get_height()
            if dest.y + line_h > rect.h:
                buff.appendleft(line)
                # FUTURE COLS ###
                if col_n < ncols:
                    dest.move_ip(col_w+col_space, -dest.y)
                    col_n += 1
                else:
                    stderr.write(WARN_BUFF)
                    return 0
                # break # FUTURE COLS DEL
                #################
            else:
                image.blit(line, dest)

                for link in line.links:
                    for _rect in line.links[link]:
                        # move rect to token's pos on image and append to links
                        _rect.move_ip(dest.topleft) # FUTURE ADD
                        # FUTURE DEL
                        #links[link].append(_rect.move(dest.topleft))
                        links[link].append(_rect)
                        # FUTURE ###
                        if link in editors: editors[link].rect = _rect
                        ############

                dest.y += line_h + spacing

        # FUTURE ###
        for editor in editors.values(): image.blit(editor.image, editor.rect)
        ############
        # FUTURE COLS ###
        self.col_n = col_n
        #################
        return 1


    def clear(self, *a):
        """
        draws background over surface_dest using self.rect and resets self._dest

        surface_dest-- the destination surface that self.image has been drawn to
        background-- the background to draw over the areas that self.image was
        drawn to on surface_dest

        returns nothing
        """
        # reset glyph
        self._dest = Rect(0, 0, 0, 0)
        self.links = defaultdict(list)
        self.col_n = 1
        self.buff = deque()
        rect = self.rect
        self.image = Surface(rect.size)
        self.image.fill(self._bkg)
        # if provided, clear a surface at glyph rect
        if a:
            surface_dest, background = a
            surface_dest.blit(background, rect, rect)


    def get_collisions(self, mpos):
        """
        get collisions between a point and the linked text on the glyph surface

        mpos-- the point to check against the linked text

        if a link collides with mpos
          returns the link collinding with mpos
        if no link collides with mpos
          returns None
        """
        editors, links, rect = self.editors, self.links, self.rect
        for link in links:
            for _rect in links[link]:
                if _rect.move(rect.topleft).collidepoint(mpos): return link



"""
    def scroll(self):
    # scrolls the surface and destination rect to accomodate text in the buffer
    # before updating
    # accepts nothing
    # returns nothing
        dest, image, spacing = self._dest, self.image, self.spacing
        h = sum(line.get_height() + spacing for line in self.buffer)
        dest.move_ip(0, -h)
        image.scroll(0, -h)
        update()
"""



import pygame
import os
from os import chdir
from os.path import dirname
from glyph import Editor, Glyph, Macros
from pygame import display
from pygame import draw
from pygame import event
from pygame.font import Font
from pygame import image
from pygame import mouse
from pygame import transform
from pygame.locals import *
pygame.init()



# path constants
DIRNAME = os.path.join(dirname(__file__), "resources",)
P_REG = os.path.join(DIRNAME, "font", "silkscreen.ttf")
P_BOLD = os.path.join(DIRNAME, "font", "silkscreen_bold.ttf")


# screen constants
SCREEN_SIZE = (800, 600)
SCREEN = display.set_mode(SCREEN_SIZE)


# colors
RED = (255, 0, 0)
GREEN = (0, 255, 0)
BLUE = (0, 0, 255)

# image constants
BKGSCREEN = image.load(os.path.join(DIRNAME, "img", "bkgscreen.tga"))
BKGSCREEN = transform.scale(BKGSCREEN, SCREEN_SIZE)
BKGSCREEN = BKGSCREEN.convert()
EDITOR_BKGSCREEN = BKGSCREEN.copy()

BLUPOT = image.load(os.path.join(DIRNAME, "img", "blu_pot.tga"))
BLUPOT = BLUPOT.convert()

REDPOT = image.load(os.path.join(DIRNAME, "img", "red_pot.tga"))
REDPOT = transform.scale(REDPOT, (16, 16))
REDPOT = REDPOT.convert()


#glyph constants
FONT = Font(os.path.join(DIRNAME, "font", "silkscreen.ttf"), 8)
DEFAULT = {
    'bkg'       : (11, 11, 11),
    'color'     : (201, 192, 187),
    'font'      : FONT,
    'spacing'   : 0, #FONT.get_linesize(),
    }


#functions
def center(surf, rect):
# centers rectangles on a specified axis on a surface
    surfrect = surf.get_rect()
    rect.x = ((surfrect.w / 2) - (rect.w / 2))
    rect.y = ((surfrect.h / 2) - (rect.h / 2))


#prepare rects and surfaces
CLIP = Rect(0, 0, 560, 420)
center(BKGSCREEN, CLIP)

EDITOR_CLIP = Rect(0, 0, 400, 300)
EDITOR_INFO_CLIP = Rect(0, 0, 400, 50)
center(BKGSCREEN, EDITOR_CLIP)
center(BKGSCREEN, EDITOR_INFO_CLIP)
EDITOR_INFO_CLIP.y -= 190

BKGSCREEN.set_clip(CLIP)
BKGSCREEN.fill((1, 1, 1))
EDITOR_BKGSCREEN.set_clip(EDITOR_INFO_CLIP)
EDITOR_BKGSCREEN.fill((1, 1, 1))
EDITOR_BKGSCREEN.set_clip(EDITOR_CLIP)
EDITOR_BKGSCREEN.fill((1, 1, 1))

CLIP.w -= 10
CLIP.h -= 10
center(BKGSCREEN, CLIP)
EDITOR_CLIP.w -= 10
EDITOR_CLIP.h -= 10
center(EDITOR_BKGSCREEN, EDITOR_CLIP)
EDITOR_INFO_CLIP.w -= 10
EDITOR_INFO_CLIP.h -= 10
center(EDITOR_BKGSCREEN, EDITOR_INFO_CLIP)
EDITOR_INFO_CLIP.y -= 190


# prepare cursors
#the default cursor
DEFAULT_CURSOR = mouse.get_cursor()


#the hand cursor
_HAND_CURSOR = (
"     XX         ",
"    X..X        ",
"    X..X        ",
"    X..X        ",
"    X..XXXXX    ",
"    X..X..X.XX  ",
" XX X..X..X.X.X ",
"X..XX.........X ",
"X...X.........X ",
" X.....X.X.X..X ",
"  X....X.X.X..X ",
"  X....X.X.X.X  ",
"   X...X.X.X.X  ",
"    X.......X   ",
"     X....X.X   ",
"     XXXXX XX   ")
_HCURS, _HMASK = pygame.cursors.compile(_HAND_CURSOR, ".", "X")
HAND_CURSOR = ((16, 16), (5, 1), _HCURS, _HMASK)


#prepare the pages of text
PAGES = {
    '0' : """welcome to the glyph demo
/n
click {link startdemo; {green; here}} to learn about glyph and the glyph mini-language
/n
click {link editor; {green; here}} to see and use the editor
/n
press ESCAPE anytime to exit this demo
""",


    'editor' : """the editor allows users to input text.  the black box below is the editor surface.  click on it to give the editor focus.  once the editor has focus, keyboard input goes to it, and will be blitted to the screen.  all keyboard input is saved as string literals.  press the escape key to exit.
""",


    'startdemo' : """glyph is a library for manipulating text and printing it to a pygame window.
/n
so what?  there are about a million pygame libraries to do that already.
/n
glyph is different for a few reasons:
1) glyph provides a within string literal mini-language for text manipulation.  you can use the mini-language to indicate what you want exactly where you want it, all right within the string literal.
/n
2) glyph provides typesetting like functionality: positioning text, wrapping text, justifying text, scrolling text, switching font, text color, background color, even inserting images into text
/n
3) glyph provides 'linked' text: text that returns a value whenever the mouse is hovering over it.  this can be used to route the user around your program, or provide tooltips.
/n
one simple object, the Glyph object, provides all this functionality.  how?  because the mini-language indicates how the Glyph object should treat text.
/n
so what is this wonderous mini-language?  click {link syntax; {green; here}} to learn more
""",


    'syntax' : """glyph syntax is modeled loosely after the LaTeX environment syntax.  appropriately, we will call text manipulation in glyph an 'environment'.  a left curly bracket, /{, indicates the beginning of an environment.  the first word following the left curly bracket declares the environment type.  after the environment type there should be a space, and then a comma seperated ilst of arguments, if there are any arguments.  a semi-colon denotes the end of the environment type declaration and argument list.  the rest of the text of this environment, until the environment ends or another is began, will be printed to the pygame window according to the instructions given by the environment type declaration and arguments.  an environment is terminated with the right curly bracket.
/n
here are some example environments and what they would produce:
/n
/space{10}glyph can /{font font//silkscreen_bold.ttf, 8; bold/} text for you.
/space{10}glyph can {font font/silkscreen_bold.ttf, 8; bold} text for you.
/n
/space{10}it can also /{color 255, 0, 0; color/} text for you.
/space{10}it can also {color 255, 0, 0; color} text for you.
/n
/space{10}it can even /{font font//silkscreen_bold.ttf, 8; nest these /{color 255, 0, 0; features/} for you/}.
/space{10}it can even {font font/silkscreen_bold.ttf, 8; nest these {color 255, 0, 0; features} for you}.
/n
this syntax is a little unwieldy, and is making it difficult to read the actual text.  to compensate for this, glyph provides macros.  click {link macros; {green; here}} to learn more about macros
""",


    'macros' : """macros are used to abbreviate the declarations and argument lists of environments.  setting macros can make the text and mini-language much easier to read and work with.  to initialize a macro, add a key-value pair to the glyph.Macro dictionary.  the key of the pair will be the abbreviation used to call the macro, and the value will be a tuple containing the environment declaration and arguments.
/n
here are some example macros:
/n
/space{10}glyph.Macro['b'] = ('font', Font(silkscreen_bold, 8))
/n
/space{10}glyph.Macro['red'] = ('color', (255, 0, 0))
/n
with the macro initialized, the abbreviation can be used in place of the full environment declaration and argument list:
/n
/space{10}glyph can /{b; bold/} text for you.
/space{10}glyph can {b; bold} text for you.
/n
/space{10}it can also /{red; color/} text for you.
/space{10}it can also {red; color} text for you.
/n
/space{10}it can even /{b; nest these /{red; features/} for you/}.
/space{10}it can even {b; nest these {red; features} for you}.
/n
macros are the recommended method of utilizing glyph's functionality.  they not only make it easier to type and read the glyph input syntax, but they also allow the client code to load resources, rather than leaving glyph attempting to find and load resources; a task that it is not designed to do and not very good at.
/n
by now you may be wondering what are the different environment declarations, and how to use them.  click {link environments; {green; here}} to learn more.
""",


    'environments' : """environments are the constructs used to manipulate how text is printed to the screen.  a Glyph object is instantiated with some default settings for font, size, color, and other parameters.  all text sent to this object will be printed to the screen according to these defaults.  if text needs to be manipulated in some way to be different from the defaults, an environment must be used.  all text inside the environment will be printed to the screen according to the instructions given by the environment declaration and argument list.
/n
these are the environment types offered in this version of glyph:
/n
/space{10}font file, size
/space{20}file is the path to the directory containing the font file
/space{20}size will be the size of the font
/n
/space{10}color R, G, B
/space{20}R, G, B are the red, green, blue values, respectively.
/n
/space{10}bkg R, G, B
/space{20}R, G, B are the red, green, blue values, respectively.
/n
/space{10}link linkname
/space{20}linkname is the name of the link, and will serve as the key for all the rects associated with the
/space{20}link
/n
NOTE: pathnames should use the slash ('//') character as separators.  glyph will convert separators to the system appropriate form.
/n
environments allow users to manipulate most of the aspects of pygame's font library, without otherwise having to write any python code.  glyph offers yet one more tool that extends the functionality of the pygame library: functions.  functions can insert images from a file or other sorts of surfaces right inside the text rendered by the font library.  to learn more about functions, click {link functions; {green; here}}.
""",


    'functions' : """functions are tools that allow special characters, images, and other surfaces to be inserted into text.  these are items that pygame's font library does not normally support, so glyph provides some special accomodations for them.  functions are denoted with the slash character: '//'.  glyph uses the slash '//' instead of the typical backslash '\\' to aovid conflicts with python or system commands.  following the slash is the function's name, then a pair of curly brackets enclosing a comma seperated argument list.  not all functions take arguments, in which event the curly brackets are empty.  there is an exception to this: inserting whitespace characters or glyph's special characters as text are also, in essence, a sort of function; for special and whitespace characters, the curly brackets should be ommitted.
/n
the following are the functions supported by the glyph minilanguage
/n
/space{10}special characters in the glyph minilanguage must be preceeded by a slash.  similarly, special
/space{10}characters like newlines are indicated with the slash in glyph
/space{20}////
/space{20}///{
/space{20}//n
/n
/space{10}//space/{spacesize/}
/space{20}spacesize is the size of the space in pixels.  this will insert blank space within the text.  the space
/space{20}will be colored to match the current environment
/n
/space{10}//img/{loc/}
/space{20}loc is the loation of the image file.  any image format supported by pygame's image library can be
/space{20}loaded.
/space{20}FUTURE: a rescale argument that will be used to rescale the image to the given dimensions.
/n
just as with environments, functions may be called from entries in the glyph.Macro dictionary.  insert the macro name, and then the item that the function would normally return.  for example:
/n
/space{10}red_pot = image.load("red_pot.tga")
/space{10}red_pot = red_pot.convert()
/space{10}glyph.Macro['red_pot'] = red_pot
/n
then call the macro using the macro name as the function name with no arguments:
/n
/space{10}to restore health, use a red potion //red_pot/{/}.
/n
the line should produce something like this:
/n
/space{10}to restore health, use a red potion /red_pot{}.
/n
one final feature glyph offers is fillable text.  To learn how to add fillable text to glyph objects, click {link fillable_text; {green; here}}.
""",

    'fillable_text' : """fillable text are editor objects inserted within the text of a glyph object.  fillable text can be used to permit users to enter data into a typeset glyph documents, and the data can be accessed from the glyph.editors attribute.  try entering some data into the fillable texts below:
/n
First Name: {editor first_name, 100; Omni}
/n
Last Name: {editor last_name, 100; Armstrong}
/n
hobbies: {editor hobbies, 200; programming, biking, gaming}
/n
fillable texts are technically environments (despite that they are not text attrtibutes but surfaces, which would generally be a function).  fillable texts are inserted with the following syntax:
/space{10}/{editor name, width; default_text/}
/space{10}'name' will the key where the editor may be accessed from glyph.editors.
/space{10}w is the the width of the fillable text.  height will be set from the surrounding line height.
/space{10}the font, color and background color of the editor are set from the default values of the glyph
/space{20}object.
/n
that's everything, and this is the end of the glyph demo.  thanks for trying it out.  glyph aims to be pygame's premier text processing library, if you have any questions, comments, concerns, or suggestions, please add them as a comment to glyph's pygame page.
"""}



class Main():
#   Example usage of Glyph

    def __init__(self):
        self.glyph = Glyph(CLIP, ncols=2, **DEFAULT)

        Macros['b'] = ('font', Font(P_BOLD, 8))
        Macros['big'] = ('font', Font(P_REG, 16))
        Macros['BIG'] = ('font', Font(P_BOLD, 16))
        Macros['red'] = ('color', RED)
        Macros['green'] = ('color', GREEN)
        Macros['bkg_blu'] = ('bkg', BLUE)
        Macros['red_pot'] = REDPOT

        self.editor = Editor(EDITOR_CLIP, **DEFAULT)
        self.editor_info = Glyph(EDITOR_INFO_CLIP, **DEFAULT)


    def start(self):
        chdir(DIRNAME)
        glyph = self.glyph
        start_editor = self.start_editor
        editor = None
        editors = glyph.editors
        glyph_rect = glyph.rect
        glyph.input(PAGES['0'], justify = 'justified')
        glyph.update()
        SCREEN.blit(BKGSCREEN, (0, 0))
        SCREEN.blit(glyph.image, glyph_rect)
        while 1:
            link = glyph.get_collisions(mouse.get_pos())
            if link: mouse.set_cursor(*HAND_CURSOR)
            else: mouse.set_cursor(*DEFAULT_CURSOR)
            for ev in event.get():
                if ev.type == MOUSEBUTTONDOWN:
                    if link:
                        if link == 'editor': start_editor()
                        elif link in editors:
                            if editor: editor.redraw() #clear the cursor
                            editor = editors[link]
                        else:
                            glyph.clear(SCREEN, BKGSCREEN)
                            glyph.input(PAGES[link], justify = 'justified')
                            glyph.update()
                            for _editor in glyph.editors.values():
                                draw.rect(glyph.image, GREEN,
                                          _editor.rect.inflate(2, 2), 1)

                if ev.type == KEYDOWN:
                    if ev.key == K_ESCAPE: exit()
                    elif editor: editor.input(ev)

            if editor: #if it has editors, glyph must be updated each loop
                cursor = editor.get_cursor()
                editor.image.fill((255, 205, 0), cursor)
                glyph.update()

            SCREEN.blit(glyph.image, glyph_rect)
            display.update()


    def start_editor(self):
        glyph = self.editor_info
        glyph_rect = glyph.rect
        glyph.input(PAGES['editor'])
        glyph.update()
        editor = self.editor
        editor_rect = editor.rect
        SCREEN.blit(EDITOR_BKGSCREEN, (0, 0))
        SCREEN.blit(glyph.image, glyph_rect)
        SCREEN.blit(editor.image, editor_rect)
        editor_focus = False
        while 1:
            mouse_pos = mouse.get_pos()
            link = glyph.get_collisions(mouse_pos)
            if link: mouse.set_cursor(*HAND_CURSOR)
            else: mouse.set_cursor(*DEFAULT_CURSOR)
            for ev in event.get():
                if ev.type == MOUSEBUTTONDOWN:
                    if link: pass

                    if editor.rect.collidepoint(mouse_pos): editor_focus = True

                    else: editor_focus = False

                if ev.type == KEYDOWN:
                    if ev.key == K_ESCAPE: exit()
                    if editor_focus == True: editor.input(ev)

            cursor = editor.get_cursor()
            editor.image.fill((255, 205, 0), cursor)

            SCREEN.blit(editor.image, editor_rect)
            display.update()




if __name__ == '__main__':
    main = Main()
    main.start()



"""
glyph is a library for manipulating text and printing it to a pygame window.

so what?  there are about a million pygame libraries to do that already.

glyph is different for a few reasons:
1) glyph provides a within string literal mini-language for text manipulation.  you can use the mini-language to indicate what you want exactly where you want it, all right within the string literal.

2) glyph provides typesetting like functionality: positioning text, wrapping text, justifying text, scrolling text, switching font, text color, background color, even inserting images into text

3) glyph provides 'linked' text: text that returns a value whenever the mouse is hovering over it.  this can be used to route the user around your program, or provide tooltips.

one simple object, the Glyph object, provides all this functionality. how? because Glyph interprets a mini-language, provided in string literals, that indicates how the Glyph object should treat text.


GLYPH MINI-LANGUAGE

syntax--
glyph syntax is modeled loosely after the LaTeX environment syntax.  appropriately, we will call text manipulation in glyph an 'environment'.  a left curly bracket, /{, indicates the beginning of an environment.  the first word following the left curly bracket declares the environment type.  after the environment type there should be a space, and then a comma seperated ilst of arguments, if there are any arguments.  a semi-colon denotes the end of the environment type declaration and argument list.  the rest of the text of this environment, until the environment ends or another is began, will be printed to the pygame window according to the instructions given by the environment type declaration and arguments.  an environment is terminated with the right curly bracket.

here are some example environments:

  glyph can {font font/silkscreen_bold.ttf, 8; bold} text for you.

  it can also {color 255, 0, 0; color} text for you.

  it can even {font font/silkscreen_bold.ttf, 8; nest these {color 255, 0, 0; features} for you}.


macros--
the glyph syntax is a little unwieldy, and is making it difficult to read the actual text.  to compensate for this, glyph provides macros.  macros are used to abbreviate the declarations and argument lists of environments.  setting macros can make the text and mini-language much easier to read and work with.  to initialize a macro, add a key-value pair to the glyph.Macro dictionary.  the key of the pair will be the abbreviation used to call the macro, and the value will be a tuple containing the environment declaration and arguments.

here are som example macros:

  glyph.Macro['b'] = ('font', Font(silkscreen_bold, 8))

  glyph.Macro['red'] = ('color', (255, 0, 0))

with the macro initialized, the abbreviation can be used in place of the full environment declaration and argument list:

  glyph can {b; bold} text for you.

  it can also {red; color} text for you.

  it can even {b; nest these {red; features} for you}.

macros are the recommended method of utilizing glyph's functionality.  they not only make it easier to type and read the glyph input syntax, but they also allow the client code to load resources, rather than leaving glyph attempting to find and load resources; a task that it is not designed to do and not very good at.


environments--
environments are the constructs used to manipulate how text is printed to the screen.  a Glyph object is instantiated with some default settings for font, size, color, and other parameters.  all text sent to this object will be printed to the screen according to these defaults.  if text needs to be manipulated in some way to be different from the defaults, an environment must be used.  all text inside the environment will be printed to the screen according to the instructions given by the environment declaration and argument list.

these are the environment types offered in this version of glyph:

  font file, size
    file is the path to the directory containing the font file
    size will be the size of the font

  color R, G, B
    R, G, B are the red, green, blue values, respectively.

  bkg R, G, B
    R, G, B are the red, green, blue values, respectively.

  link linkname
    linkname is the name of the link, and will serve as the key for all the rects associated with the link

NOTE: pathnames should use the slash ('//') character as separators.  glyph will convert separators to the system appropriate form.

environments allow users to manipulate most of the aspects of pygame's font library, without otherwise having to write any python code.


functions--
functions can insert images from a file or other sorts of surfaces right inside the text rendered by the font library.  functions are tools that allow special characters, images, and other surfaces to be inserted into text.  these are items that pygame's font library does not normally support, so glyph provides some special accomodations for them.  functions are denoted with the slash character: '//'.  glyph uses the slash '//' instead of the typical backslash '\\' to aovid conflicts with python or system commands.  following the slash is the function's name, then a pair of curly brackets enclosing a comma seperated argument list.  not all functions take arguments, in which event the curly brackets are empty.  there is an exception to this: inserting whitespace characters or glyph's special characters as text are also, in essence, a sort of function; for special and whitespace characters, the curly brackets should be ommitted.

the following are the functions supported by the glyph minilanguage

  special characters in the glyph minilanguage must be preceeded by a slash.  similarly, special characters like newlines are indicated with the slash in glyph
    //
    /{
    /n

  /space{spacesize}
    spacesize is the size of the space in pixels.  this will insert blank space within the text.  the space will be colored to match the current environment

  /img{loc}
    loc is the loation of the image file.  any image format supported by pygame's image library can be loaded.
    FUTURE: a rescale argument that will be used to rescale the image to the given dimensions.

just as with environments, functions may be called from entries in the glyph.Macro dictionary.  insert the macro name, and then the item that the function would normally return.  for example:

  red_pot = image.load("red_pot.tga")
  red_pot = red_pot.convert()
  glyph.Macro['red_pot'] = red_pot

then call the macro using the macro name as the function name with no arguments:

  to restore health, use a red potion /red_pot{}.


EXAMPLE

see the example folder, provided with the source distribution, for an example script.  Running the example will cover the same topics introduced above, but will show the results of the examples.
"""




from .glyph import Glyph, GlyphGroup, Macros
from .editor import Editor, EditorGroup



#!/usr/bin/env python

from distutils.core import setup

setup(name='glyph',
      version='2.6.5b',
      author='Chandler Armstrong',
      author_email='omni.armstrong@gmail.com',
      url='http://code.google.com/p/glyph/',
      description='Pygame typesetting library',
      download_url='http://code.google.com/p/glyph/downloads/list',
      classifiers=['Development Status :: 5 - Production/Stable',
                   'License :: OSI Approved :: GNU General Public License (GPL)',
                   'Topic :: Software Development :: Libraries :: pygame',
                   'Topic :: Text Processing'],
      packages=['glyph'],
      requires=['pygame(>=1.9.1)'],
      provides=['glyph'])
